from pydantic import BaseModel, AfterValidator
from typing import Annotated, Dict


def check_corresponding_keys(dictionary):
    for key, v in dictionary.items():
        if key == v.key:
            continue
        raise ValueError(f"mismatch of {key=} and {v.key=}")
    return dictionary


class EntrypointConfig(BaseModel):
    key: str
    name: str
    caption: str
    args: dict[str, str | bool | int]


Entrypoints = Annotated[dict[str, EntrypointConfig], AfterValidator(check_corresponding_keys)]


class EntrypointsConfig(BaseModel):
    entrypoints: Entrypoints
    default_entrypoint_key: str
    warmup: bool



import logging
from functools import lru_cache

from .entrypoints_config import EntrypointsConfig
from .entrypoints import create_entrypoint


logger = logging.getLogger(__name__)


class EntrypointsAccessor:
    def __init__(self, cfg: EntrypointsConfig):
        self.cfg = cfg

    def _warmup_if_needed(self):
        if not self.cfg.warmup:
            return
        assert self.cfg.default_entrypoint_key in self.cfg.entrypoints
        for ec_key in self.cfg.entrypoints.keys():
            self.get(ec_key)

    @lru_cache(None)
    def get_default(self):
        default_key = self.cfg.default_entrypoint_key
        entrypoint = self.get(default_key)
        return entrypoint

    @lru_cache(None)
    def get(self, entrypoint_key: str):
        ep = self.cfg.entrypoints.get(entrypoint_key)
        if ep is None:
            default_key = self.cfg.default_entrypoint_key
            logger.warning(f"Not found entrypoint for key={entrypoint_key}, fallback to {default_key}")
            ep = self.cfg.entrypoints[default_key]
        entrypoint = create_entrypoint(ep.name, ep.args)
        return entrypoint

    def __getitem__(self, key):
        return self.get(key)

    def clear(self):
        self.get.cache_clear()



__version__ = "1.11.6"

from .base_chat import AbstractEntryPoint
from .entrypoints import (
    OpenRouterEntryPoint,
    AiriChatEntryPoint,
    YandexGPTEntryPoint,
    GigaChatCensoredEntryPoint,
    GigaChatEntryPoint,
    GigaPlusEntryPoint,
    GigaMaxEntryPoint,
    GigaMax2EntryPoint,
    GigaMax2SberdevicesEntryPoint,
    FusionBrainEntrypoint,
)

from .entrypoints_accessor import (
    create_entrypoint,
    EntrypointsAccessor,
)

from .entrypoints_config import EntrypointsConfig



from contextlib import contextmanager
from functools import partial
from typing import Literal
from collections.abc import Callable


OnError = Literal["ignore", "warn", "fail"]
ErrorHandler = Callable[[Exception], ...]


def _on_exception(error_handler: ErrorHandler):
    try:
        yield
    except Exception as e:
        error_handler(e)


def make_on_exception(on_error: OnError, logger, err_msg: str):
    error_handler = _make_error_handler(on_error, logger, err_msg)
    res = contextmanager(partial(_on_exception, error_handler=error_handler))
    return res


def _make_error_handler(on_error: OnError, logger, err_msg: str) -> ErrorHandler:
    if on_error == "ignore":
        return _error_ignorer
    elif on_error == "warn":
        return partial(_error_warner, logger=logger, err_msg=err_msg)
    elif on_error == "fail":
        return _error_raiser
    else:
        raise ValueError(f"Expected OnError value, found: {on_error}")


def _error_ignorer(ex):
    pass


def _error_warner(ex, logger, err_msg):
    if logger:
        logger.warning(err_msg)
    else:
        warnings.warn(err_msg)


def _error_raiser(ex):
    raise ex



import base64
from time import sleep
from typing import List, Dict

from openai import OpenAI, DefaultHttpxClient
from loguru import logger

from .base_chat import AbstractEntryPoint
from .io_sugar import make_on_exception, OnError


class OpenRouterEntryPoint(AbstractEntryPoint):
    def __init__(
        self,
        model_id: str,
        api_key: str,
        base_url="https://openrouter.ai/api/v1",
        emb_dim: int = 1024,
        providers: list[str] = [],
        warmup: bool = False,
        on_error: OnError = "ignore",
        verify: bool = True,
    ) -> None:
        self._model = OpenAI(base_url=base_url, api_key=api_key, http_client=DefaultHttpxClient(verify=verify))
        self.model_id: str = model_id
        self.extra_body: dict[str, dict[str, str]] = {"provider": {"order": providers}}
        self._DIM: int = emb_dim
        self._ZEROS: list[float] = [0 for _ in range(self._DIM)]
        self._ERROR_MESSAGE: str = ""
        if warmup:
            self.warmup()
        self.on_exception = make_on_exception(on_error, logger, err_msg=f"No response from {model_id}")

    def __call__(self) -> OpenAI:
        return self._model

    def get_response(self, sentence: str, **kwargs) -> str:
        with self.on_exception():
            payload = self.create_payload(system_prompt="", user_prompt=sentence)
            return (
                self._model.chat.completions.create(
                    model=self.model_id, messages=payload, extra_body=self.extra_body, **kwargs
                )
                .choices[0]
                .message.content
            )
        return self._ERROR_MESSAGE

    def get_instruct_response(self, sentence: str, **kwargs) -> str:
        with self.on_exception():
            return self._model.completions.create(model=self.model_id, prompt=sentence, **kwargs).choices[0].text
        return self._ERROR_MESSAGE

    def get_image_response(self, bytesimage: bytes, sentences: str, mimetype: str = "image/jpeg") -> str:
        with self.on_exception():
            encoded_image = base64.b64encode(bytesimage).decode("utf-8")
            payload = self.create_image_payload(
                system_prompt="", user_prompt=sentences, image_encoded=encoded_image, mimetype=mimetype
            )
            return (
                self._model.chat.completions.create(
                    model=self.model_id,
                    messages=payload,
                    extra_body=self.extra_body,
                )
                .choices[0]
                .message.content
            )
        return self._ERROR_MESSAGE

    def get_response_by_payload(self, payload: list[dict[str, str]], **kwargs) -> str:
        """payload: [{"role": "system", "content": system}, {"role": "user", "content": replica}]"""
        with self.on_exception():
            return (
                self._model.chat.completions.create(
                    model=self.model_id, messages=payload, extra_body=self.extra_body, **kwargs
                )
                .choices[0]
                .message.content
            )
        return self._ERROR_MESSAGE

    def get_embedding(self, sentence: str) -> list[float]:
        with self.on_exception():
            return self._model.embeddings.create(model=self.model_id, input=[sentence]).data[0].embedding
        return self._ZEROS

    def get_embeddings(self, sentences: list[str], request_limit=50) -> list[list[float]]:
        embeddings: list[list[float]] | None = None
        counter: int = 0
        while embeddings is None and counter < request_limit:
            with self.on_exception():
                embeddings = [
                    item.embedding for item in self._model.embeddings.create(model=self.model_id, input=sentences).data
                ]
                break
            sleep(0.1)
            counter += 1
        if embeddings is not None:
            return embeddings
        return [self._ZEROS for _ in sentences]

    def __repr__(self):
        class_name = type(self).__name__
        return f"{class_name}(model_id={self.model_id})"

    def __str__(self):
        return repr(self)



import logging
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from itertools import chain
from typing import Any

import tiktoken
from httpx import NetworkError


logger = logging.getLogger(__name__)


class AbstractEntryPoint(ABC):
    @abstractmethod
    def __call__(self) -> Any:
        pass

    @abstractmethod
    def get_response(self, sentence: str) -> str:
        pass

    @abstractmethod
    def get_response_by_payload(self, payload: list[dict[str, str]]) -> str:
        pass

    @abstractmethod
    def get_embedding(self, sentence: str) -> list[float]:
        pass

    @abstractmethod
    def get_embeddings(self, sentences: list[str], request_limit: int = 50) -> list[list[float]]:
        pass

    def decorate_with(self, decorator) -> "AbstractEntryPoint":
        return EntryPointFacade(self, decorator)

    def get_response_with_retries(self, prompt: str, retries: int = 3) -> str:
        response = self.get_response(prompt)
        retry_count = 0
        if response == "":
            while retry_count < retries:
                time.sleep(5)
                logger.warning(f"Response is empty, attempt {retry_count}")
                retry_count += 1
                response = self.get_response(prompt)
                if response != "":
                    break
        if response == "":
            logger.error("Response is empty!")
        if retry_count > 0:
            logger.info(f"Non-empty answer after retry (count={retry_count})")
        return response

    def get_responses(self, sentences: list[str], max_workers: int = 2) -> list[str]:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(self.get_response, sentence) for sentence in sentences]
            responses = [future.result() for future in futures]
        return responses

    def get_responses_by_payload(self, payloads: list[list[dict[str, str]]], max_workers: int = 2) -> list[str]:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(self.get_response, payload) for payload in payloads]
            responses = [future.result() for future in futures]
        return responses

    def get_more_embeddings(self, sentences: list[str], batch_size: int = 2, max_workers: int = 4) -> list[list[float]]:
        batches: list[list[str]] = self.make_batches(sentences, size=batch_size)
        if max_workers == 1:
            emb_batches = [self.get_embeddings(batch) for batch in batches]
        else:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = [executor.submit(self.get_embeddings, batch) for batch in batches]
                emb_batches = [future.result() for future in futures]
        return list(chain.from_iterable(emb_batches))

    @staticmethod
    def count_tokens(sentences: list[str]) -> list[int]:
        encoding = tiktoken.get_encoding("cl100k_base")
        return [len(encoding.encode(sentence)) for sentence in sentences]

    @staticmethod
    def make_batches(items: list, size: int = 500) -> list[list[str]]:
        slices = [(i * size, (i + 1) * size) for i in range(len(items) // size + 1)]
        return [items[st:ed] for st, ed in slices]

    def warmup(self) -> None:
        response = self.get_response("Прогрев")
        if not response or not sum(self.get_embedding("Прогрев")):
            raise NetworkError("Нет доступа к ллм!")

    @staticmethod
    def create_payload(system_prompt: str, user_prompt: str) -> list[dict[str, str]]:
        return [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_prompt},
        ]

    @staticmethod
    def create_image_payload(system_prompt: str, user_prompt: str, image_encoded: str, mimetype="image/jpeg"):
        image_content = [
            {"type": "text", "text": user_prompt},
            {"type": "image_url", "image_url": {"url": f"data:{mimetype};base64,{image_encoded}"}},
        ]
        return [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": image_content},
        ]


class EntryPointFacade(AbstractEntryPoint):
    def __init__(self, base: AbstractEntryPoint, decorator):
        self._base = base

        # with basic class (not ABC) it's possible to override method directly
        # like this: `self.get_response = decorator(self._base.get_response)`
        self._get_response = decorator(self._base.get_response)
        self._get_response_by_payload = decorator(self._base.get_response_by_payload)
        self._get_embedding = decorator(self._base.get_embedding)
        self._get_embeddings = decorator(self._base.get_embeddings)

    def __call__(self) -> Any:
        return self._base.__call__()

    def get_response(self, sentence: str) -> str:
        return self._get_response(sentence)

    def get_response_by_payload(self, payload: list[dict[str, str]]) -> str:
        return self._get_response_by_payload(payload)

    def get_embedding(self, sentence: str) -> list[float]:
        return self._get_embedding(sentence)

    def get_embeddings(self, sentences: list[str], request_limit: int = 50) -> list[list[float]]:
        return self._get_embeddings(sentences, request_limit=request_limit)



from .open_router_entrypoint import OpenRouterEntryPoint
from .airi_entrypoint import AiriChatEntryPoint
from .yandex_gpt_entrypoint import YandexGPTEntryPoint
from .fusion_brain_entrypoint import FusionBrainEntrypoint
from .gigachat_entrypoint import (
    GigaChatCensoredEntryPoint,
    GigaChatEntryPoint,
    GigaPlusEntryPoint,
    GigaMaxEntryPoint,
    GigaMax2EntryPoint,
    GigaMax2SberdevicesEntryPoint,
)

ENTRYPOINTS: dict[str, str] = {
    "airi": AiriChatEntryPoint,
    "giga": GigaChatEntryPoint,
    "giga-max": GigaMaxEntryPoint,
    "giga-max-2": GigaMax2EntryPoint,
    "giga-plus": GigaPlusEntryPoint,
    "giga-censored": GigaChatCensoredEntryPoint,
    "open-router": OpenRouterEntryPoint,
    "yandex": YandexGPTEntryPoint,
    "fusion-brain": FusionBrainEntrypoint,
    "giga-max-2-sberdevices": GigaMax2SberdevicesEntryPoint,
}


def create_entrypoint(entrypoint_name: str, entrypoint_args: dict[str, any]):
    entrypoint_class = ENTRYPOINTS.get(entrypoint_name)
    if entrypoint_class is None:
        err = f"Not found entrypoint for entrypoint_name={entrypoint_name}"
        raise ValueError(err)
    return entrypoint_class(**entrypoint_args)



import logging
from urllib.parse import urljoin

import requests

from .models import ServiceUnavailableException, UnsupportedModelException
from .open_router_entrypoint import OpenRouterEntryPoint


logger = logging.getLogger(__name__)


class AiriChatEntryPoint(OpenRouterEntryPoint):
    def __init__(
        self,
        model_id: str,
        api_key="hello_darkness",
        base_url="http://192.168.24.161:30050",
        emb_dim: int = 1024,
        verify: bool = True,
    ) -> None:
        server_url = urljoin(base_url, "v1")
        self.models_url = urljoin(base_url, "v1/models")
        self.api_key = api_key
        self.verify = verify
        self.check_current(model_id=model_id)
        super().__init__(model_id=model_id, base_url=server_url, api_key=api_key, emb_dim=emb_dim, verify=verify)

    def get_url(self):
        return self.models_url

    def check_current(self, model_id) -> None:
        model_list = self.get_available_models()

        if model_id not in model_list:
            logger.error("Модель '%s' не найдена. Доступные модели: %s", model_id, model_list)
            raise UnsupportedModelException(model_id, model_list)
        else:
            logger.info("Модель '%s' успешно найдена среди доступных.", model_id)

    def get_available_models(self) -> list[str]:
        response = requests.get(
            self.models_url, verify=self.verify, headers={"Authorization": f"Bearer {self.api_key}"}
        )
        available_models = []
        if response.status_code == 200:
            models = response.json()
            for m in models["data"]:
                available_models.append(m["id"])
            logger.info("Доступные модели: %s", available_models)
            return available_models
        else:
            raise ServiceUnavailableException()



import logging
from base64 import b64encode
from time import sleep
from typing import Literal

from gigachat import GigaChat
from gigachat._types import FileTypes
from gigachat.models import Embedding, UploadedFile

from .base_chat import AbstractEntryPoint
from .io_sugar import make_on_exception, OnError


logger = logging.getLogger(__name__)


def _make_kwargs_auth(user: str, password: str, client_id: str, client_secret: str):
    if user and password:
        return dict(user=user, password=password)
    if client_id and client_secret:
        creds = b64encode(f"{client_id}:{client_secret}".encode("utf-8")).decode("utf-8")
        return dict(credentials=creds)
    raise ValueError(f"Unexpected combination of auth args: {user=}, {password=}, {client_id=}, {client_secret=}")


BASE_URL_DEFAULT = "https://gigachat.devices.sberbank.ru/api/v1"


class GigaChatEntryPoint(AbstractEntryPoint):
    def __init__(
        self,
        user: str | None = None,
        password: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        model_id: str = "GigaChat-Pro",
        warmup: bool = False,
        temperature: float = 0.0,
        base_url: str = BASE_URL_DEFAULT,
        profanity_check: bool = False,
        on_error: OnError = "ignore",
    ) -> None:
        kwargs_pre = dict(
            base_url=base_url,
            scope="GIGACHAT_API_CORP",
            model=model_id,
            verify_ssl_certs=False,
            profanity_check=False,
        )
        kwargs_auth = _make_kwargs_auth(user, password, client_id, client_secret)
        kwargs = kwargs_pre | kwargs_auth
        self._base_url = base_url
        self._model_id = model_id
        self._model = GigaChat(**kwargs)
        self._DIM: int = 1024
        self._ZEROS: list[float] = [0.0 for _ in range(self._DIM)]
        self._ERROR_MESSAGE: str = ""
        self.temperature = temperature
        if warmup:
            self.warmup()
        self.on_exception = make_on_exception(on_error, logger, err_msg=f"No response from {model_id}")

    def __call__(self) -> GigaChat:
        return self._model

    def get_response(self, sentence: str) -> str:
        with self.on_exception():
            return self._model.chat(sentence).choices[0].message.content
        return self._ERROR_MESSAGE

    def get_response_by_payload(self, payload: list[dict[str, str]]) -> str:
        """payload: [{"role": "system", "content": system}, {"role": "user", "content": replica}]"""
        with self.on_exception():
            return self._model.chat({"messages": payload, "temperature": self.temperature}).choices[0].message.content
        return self._ERROR_MESSAGE

    def get_embedding(self, sentence: str) -> list[float]:
        with self.on_exception():
            return self._model.embeddings([sentence]).data[0].embedding
        return self._ZEROS

    def get_embeddings(self, sentences: list[str], request_limit: int=50) -> list[list[float]]:
        embeddings: list[list[float]] | None = None
        counter: int = 0
        while embeddings is None and counter < request_limit:
            with self.on_exception():
                items: list[Embedding] = self._model.embeddings(sentences).data
                embeddings = [item.embedding for item in items]
                break
            sleep(0.1)
            counter += 1
        if embeddings is not None:
            return embeddings
        return [self._ZEROS for _ in sentences]

    def upload_file(
        self,
        file: FileTypes,
        purpose: Literal["general", "assistant"] = "general",
    ) -> UploadedFile:
        return self._model.upload_file(file, purpose)

    def __repr__(self):
        class_name = type(self).__name__
        url = self._base_url
        url_info = "" if BASE_URL_DEFAULT == url else f", url={url}"
        model_info = f"model_id={self._model_id}"
        return f"{class_name}({model_info}{url_info})"

    def __str__(self):
        return repr(self)


def _G(**kwargs_base):
    def instantiate_gigachat(**kwargs):
        return GigaChatEntryPoint(**kwargs_base, **kwargs)

    return instantiate_gigachat


GigaChatCensoredEntryPoint = _G(model_id="GigaChat-Pro", profanity_check=True)
GigaMaxEntryPoint = _G(model_id="GigaChat-Max")
GigaMax2EntryPoint = _G(model_id="GigaChat-2-Max")
GigaPlusEntryPoint = _G(model_id="GigaChat-Plus")
GigaMax2SberdevicesEntryPoint = _G(model_id="GigaChat-2-Max", base_url="https://gigachat.sberdevices.ru/v1")
